# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
"""Python 3.14 compatibility patches for Hydra.

This module contains patches to fix compatibility issues with Hydra on Python 3.14.
Must be imported before any Hydra imports occur.
"""

import sys


def apply_hydra_argparse_patch() -> None:  # pragma: no cover
    """Apply Python 3.14 compatibility patch for Hydra's argparse usage.

    Python 3.14 added validation in argparse._check_help that checks if '%'
    is in the help string. Hydra's LazyCompletionHelp object doesn't support
    the 'in' operator, causing a TypeError during argument parser creation.

    This patch temporarily disables the validation during Hydra's parser
    creation, then immediately restores it. This is safe because:
    1. The validation is purely for catching developer errors early
    2. It's new in Python 3.14 (didn't exist in earlier versions)
    3. Hydra's help strings work correctly when actually displayed
    4. We restore normal behavior immediately after
    """
    if sys.version_info < (3, 14):
        return

    try:
        import argparse
        from functools import wraps

        from hydra._internal import utils as hydra_utils

        original_get_args_parser = hydra_utils.get_args_parser
        original_check_help = argparse.ArgumentParser._check_help

        @wraps(original_get_args_parser)
        def patched_get_args_parser():  # type: ignore
            """Patched version that disables help validation during parser creation."""
            argparse.ArgumentParser._check_help = lambda self, action: None
            try:
                return original_get_args_parser()
            finally:
                argparse.ArgumentParser._check_help = original_check_help

        hydra_utils.get_args_parser = patched_get_args_parser

        if "hydra.main" in sys.modules:
            sys.modules["hydra.main"].get_args_parser = patched_get_args_parser

    except (ImportError, AttributeError):
        pass



# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
# pyright: reportUnnecessaryTypeIgnoreComment=false
import functools
import inspect
from typing import Any, Callable, TypeVar, Union, cast

import pydantic as _pyd

_T = TypeVar("_T", bound=Callable[..., Any])

__all__ = ["validates_with_pydantic"]


if _pyd.__version__ >= "2.0":  # pragma: no cover
    _default_parser = _pyd.validate_call(
        config={"arbitrary_types_allowed": True}, validate_return=False  # type: ignore
    )
else:  # pragma: no cover
    _default_parser = _pyd.validate_arguments(
        config={"arbitrary_types_allowed": True, "validate_return": False}  # type: ignore
    )


def _constructor_as_fn(cls: Any) -> Any:
    """Makes a shim around a class constructor so that it is compatible with pydantic validation.

    Notes
    -----
    `pydantic.validate_call` mishandles class constructors; it expects that
    `cls`/`self` should be passed explicitly to the constructor. This shim
    corrects that.
    """

    @functools.wraps(cls)
    def wrapper_function(*args, **kwargs):
        return cls(*args, **kwargs)

    annotations = getattr(cls, "__annotations__", {})

    # In a case like:
    # class A:
    #   x: int
    #   def __init__(self, y: int): ...
    #
    #  y will not be in __annotations__ but it should be in the signature,
    #  so we add it to the annotations.

    sig = inspect.signature(cls)
    for p, v in sig.parameters.items():
        if p not in annotations:
            annotations[p] = v.annotation
    wrapper_function.__annotations__ = annotations

    return wrapper_function


def _get_signature(x: Any) -> Union[None, inspect.Signature]:
    try:
        return inspect.signature(x)
    except Exception:
        return None


def pydantic_parser(target: _T, *, parser: Callable[[_T], _T] = _default_parser) -> _T:
    """A target-wrapper that adds pydantic parsing to the target.

    This can be passed to `instantiate` as a `_target_wrapper_` to add pydantic parsing
    to the (recursive) instantiation of the target.

    Parameters
    ----------
    target : Callable

    parser : Type[pydantic.validate_arguments], optional
        A configured instance of pydantic's validation decorator.

        The default validator that we provide specifies:
           - arbitrary_types_allowed: True

    Examples
    --------
    .. code-block:: python

       from hydra_zen import builds, instantiate
       from hydra_zen.third_party.pydantic import pydantic_parser

       from pydantic import PositiveInt

       def f(x: PositiveInt): return x

       good_conf = builds(f, x=10)
       bad_conf = builds(f, x=-3)

    >>> instantiate(good_conf, _target_wrapper_=pydantic_parser)
    10
    >>> instantiate(bad_conf, _target_wrapper_=pydantic_parser)
    ValidationError: 1 validation error for f (...)

    This also enables type conversion / parsing. E.g. Hydra can
    only produce lists from the CLI, but this parsing layer can
    convert them based on the annotated type. (Note: this only
    works for pydantic v2 and higher.)

    >>> def g(x: tuple): return x
    >>> conf = builds(g, x=[1, 2, 3])
    >>> instantiate(conf, _target_wrapper_=pydantic_parser)
    (1, 2, 3)
    """
    if inspect.isbuiltin(target):
        return cast(_T, target)

    if isinstance(target, type) and issubclass(target, _pyd.BaseModel):
        # this already applies pydantic parsing
        return cast(_T, target)

    if not (_get_signature(target)):
        return cast(_T, target)

    if inspect.isclass(target):
        return cast(_T, parser(_constructor_as_fn(target)))

    return parser(target)


def validates_with_pydantic(
    obj: _T, *, validator: Callable[[_T], _T] = _default_parser
) -> _T:
    """
    .. deprecated:: 0.13.0

       Use `hydra_zen.third_party.pydantic.pydantic_parser` instead.
    """
    return pydantic_parser(obj, parser=validator)



# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import inspect
import warnings
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import MISSING, field as _field, is_dataclass
from keyword import iskeyword
from typing import Any, Callable, Final, Optional, TypeVar, Union, cast, overload

from omegaconf import II, DictConfig, ListConfig
from typing_extensions import Literal, TypeGuard

from hydra_zen._compatibility import HYDRA_SUPPORTS_OBJECT_CONVERT
from hydra_zen.errors import HydraZenValidationError
from hydra_zen.typing import DataclassOptions, ZenConvert
from hydra_zen.typing._implementations import (
    DEFAULT_DATACLASS_OPTIONS,
    UNSUPPORTED_DATACLASS_OPTIONS,
    AllConvert,
    Field,
    InterpStr,
    StrictDataclassOptions,
    convert_types,
)

COMMON_MODULES_WITH_OBFUSCATED_IMPORTS: tuple[str, ...] = (
    "random",
    "numpy",
    "numpy.random",
    "jax.numpy",
    "jax",
    "torch",
)
UNKNOWN_NAME: Final[str] = "<unknown>"
KNOWN_MUTABLE_TYPES: set[
    Union[type[list[Any]], type[dict[Any, Any]], type[set[Any]]]
] = {list, dict, set}

T = TypeVar("T")


# The typeshed definition of `field` has an inaccurate annotation:
#  https://github.com/python/typeshed/blob/b9e1d7d522fe90b98e07d43a764bbe60216bc2c4/stdlib/dataclasses.pyi#L109
# This makes it impossible for `make_dataclass` to by type-correct in the eyes of
# static checkers. See https://github.com/microsoft/pyright/issues/1680 for discussion.
#
# We happen to make rather heavy use of `make_dataclass`, thus we..*sigh*.. we provide
# our own overloads for `field`.
@overload  # `default` and `default_factory` are optional and mutually exclusive.
def field(
    *,
    default: Any,
    init: bool = ...,
    repr: bool = ...,
    hash: Optional[bool] = ...,
    compare: bool = ...,
    metadata: Optional[Mapping[Any, Any]] = ...,
) -> Field[Any]:  # pragma: no cover
    ...


@overload
def field(
    *,
    default_factory: Callable[[], Any],
    init: bool = ...,
    repr: bool = ...,
    hash: Optional[bool] = ...,
    compare: bool = ...,
    metadata: Optional[Mapping[Any, Any]] = ...,
) -> Field[Any]:  # pragma: no cover
    ...


def field(
    *,
    default: Any = MISSING,
    default_factory: Union[Callable[[], Any], Any] = MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
) -> Field[Any]:
    if default is MISSING:
        return cast(
            Field[Any],
            _field(
                default_factory=default_factory,
                init=init,
                repr=repr,
                hash=hash,
                compare=compare,
                metadata=metadata,
            ),
        )
    else:
        return cast(
            Field[Any],
            _field(
                default=default,
                init=init,
                repr=repr,
                hash=hash,
                compare=compare,
                metadata=metadata,
            ),
        )


def safe_name(obj: Any, repr_allowed: bool = True) -> str:
    """Tries to get a descriptive name for an object. Returns '<unknown>`
    instead of raising - useful for writing descriptive/dafe error messages."""

    if hasattr(obj, "__name__"):
        return obj.__name__.replace("<lambda>", "lambda")

    if repr_allowed and hasattr(obj, "__repr__"):
        return repr(obj).replace("<lambda>", "lambda")

    return UNKNOWN_NAME


def is_classmethod(obj: Any) -> bool:
    """
    https://stackoverflow.com/a/19228282/6592114

    Credit to: Martijn Pieters
    License: CC BY-SA 4.0 (free to copy/redistribute/remix/transform)"""

    if not inspect.ismethod(obj):
        return False

    bound_to = getattr(obj, "__self__", None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = safe_name(obj)

    if name == UNKNOWN_NAME:  # pragma: no cover
        return False

    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False  # pragma: no cover


def building_error_prefix(target: Any) -> str:
    return f"Building: {safe_name(target)} ..\n"


NoneType = type(None)


def is_interpolated_string(x: Any) -> TypeGuard[InterpStr]:
    # This is only a necessary check – not a sufficient one – that `x`
    # is a valid interpolated string. We do not verify that it rigorously
    # satisfies omegaconf's grammar
    return isinstance(x, str) and len(x) > 3 and x.startswith("${") and x.endswith("}")


def check_suspicious_interpolations(
    validated_wrappers: Sequence[Any], zen_meta: Mapping[str, Any], target: Any
):
    """Looks for patterns among zen_meta fields and interpolated fields in
    wrappers. Relative interpolations pointing to the wrong level will produce
    a warning"""
    for _w in validated_wrappers:
        if is_interpolated_string(_w):
            _lvl = _w.count(".")  # level of relative-interp
            _field_name = _w.replace(".", "")[2:-1]
            if (
                _lvl
                and _field_name in zen_meta
                and _lvl != (1 if len(validated_wrappers) == 1 else 2)
            ):
                _expected = II(
                    "." * (1 if len(validated_wrappers) == 1 else 2) + _field_name
                )

                warnings.warn(
                    building_error_prefix(target)
                    + f"A zen-wrapper is specified via the interpolated field, {_w},"
                    f" along with the meta-field name {_field_name}, however it "
                    f"appears to point to the wrong level. It is likely you should "
                    f"change {_w} to {_expected}"
                )
                yield _expected


def valid_defaults_list(hydra_defaults: Any) -> bool:
    """
    Raises
    ------
    HydraZenValidationError: Duplicate _self_ entries"""
    if not isinstance(hydra_defaults, (list, ListConfig)):
        return False

    has_self = False
    for item in hydra_defaults:
        if item == "_self_":
            if not has_self:
                has_self = True
                continue
            raise HydraZenValidationError(
                "`hydra_defaults` cannot have more than one '_self_' entry"
            )

        if isinstance(item, (dict, DictConfig)):
            for k, v in item.items():
                if not isinstance(k, str):
                    return False

                if (
                    not isinstance(v, (str, list, ListConfig))
                    and v is not None
                    and v != MISSING
                ):
                    return False
        elif isinstance(item, str):
            continue
        elif is_dataclass(item):
            # no validation here
            continue
        else:
            return False

    if not has_self:
        warnings.warn(
            "Defaults list is missing `_self_`. See https://hydra.cc/docs/upgrades/1.0_to_1.1/default_composition_order for more information",
            category=UserWarning,
        )
    return True


def merge_settings(
    user_settings: Optional[ZenConvert], default_settings: AllConvert
) -> AllConvert:
    """Merges settings as `default_settings.update(user_settings)`"""
    if user_settings is not None and not isinstance(user_settings, Mapping):
        raise TypeError(
            f"`zen_convert` must be None or Mapping[str, Any] (e.g. dict). Got {user_settings}"
        )
    settings = default_settings.copy()
    if user_settings:
        for k, v in user_settings.items():
            if k not in convert_types:
                raise ValueError(
                    f"The key `{k}` is not a valid zen_convert setting. The available settings are: {', '.join(sorted(convert_types))}"
                )
            if not isinstance(v, convert_types[k]):
                raise TypeError(
                    f"Setting {k}={v} specified a value of the wrong type. Expected type: {convert_types[k].__name__}"
                )
            settings[k] = v
    return settings


_DATACLASS_OPTION_KEYS: frozenset[str] = (
    DataclassOptions.__required_keys__ | DataclassOptions.__optional_keys__
)

_STRICT_DATACLASS_OPTION_KEYS: frozenset[str] = (
    StrictDataclassOptions.__required_keys__ | StrictDataclassOptions.__optional_keys__
)
_STRICT_DATACLASS_OPTION_KEYS.copy()


def parse_dataclass_options(
    options: Mapping[str, Any], include_module: bool = True
) -> DataclassOptions:
    """
    Ensures `options` adheres to `DataclassOptions` and merges hydra-zen defaults
    for missing options.

    All valid `@dataclass`/`make_dataclass` options are supported, even for features
    introduced in later versions of Python. This function will remove valid options
    that are not supported for by the current Python version.

    Parameters
    ----------
    options : Mapping[str, Any]
        User-specified options for `zen_dataclass` to be validated.

    Returns
    -------
    DataclassOptions

    Examples
    --------
    >>> parse_dataclass_options({})
    {'unsafe_hash': True}

    >>> parse_dataclass_options({"unsafe_hash": False, "cls_name": "Foo"})
    {'unsafe_hash': False, 'cls_name': 'Foo'}

    >>> parse_dataclass_options({"moo": 1})
    ValueError: moo is not a valid dataclass option.

    Options that are supported by `make_dataclass` for later versions of
    Python are ignored/removed automatically by this function. E.g. the following
    Python 3.10+ option has the following behavior in Python 3.9:

    >>> parse_dataclass_options({"slots": False})
    {'unsafe_hash': True}
    """
    if not isinstance(options, Mapping):
        raise ValueError(
            f"`zen_dataclass_options` is expected to be `None` or dict[str, bool]. Got "
            f"{options} (type: {type(options)})."
        )

    merged = DEFAULT_DATACLASS_OPTIONS.copy()

    for name, val in options.items():
        if name in UNSUPPORTED_DATACLASS_OPTIONS:
            continue
        elif name not in _DATACLASS_OPTION_KEYS:
            raise ValueError(f"{name} is not a valid dataclass option.")

        if name == "module":
            if val is not None and (
                not isinstance(val, str)
                or not all(
                    v.isidentifier() and not iskeyword(v) for v in val.split(".")
                )
            ):
                raise ValueError(
                    f"dataclass option `{name}` must be a valid module name, got {val}"
                )

        elif name == "cls_name":
            if val is not None and (not isinstance(val, str) or not val.isidentifier()):
                raise ValueError(
                    f"dataclass option `{name}` must be a valid identifier, got {val}"
                )
        elif name == "bases":
            if not isinstance(val, Iterable) or any(
                not (is_dataclass(_b) and isinstance(_b, type)) for _b in val
            ):
                raise TypeError(
                    f"dataclass option `{name}` must be a tuple of dataclass types"
                )
        elif name == "namespace":
            if not isinstance(val, Mapping) or any(
                not isinstance(v, str) or not v.isidentifier() for v in val
            ):
                raise ValueError(
                    f"dataclass option `{name}` must be a mapping with string-valued keys "
                    f"that are valid identifiers. Got {val}."
                )
        elif name == "target":
            if not isinstance(val, str) or not all(
                x.isidentifier() for x in val.split(".")
            ):
                raise TypeError(
                    f"dataclass option `target` must be a string and an import path, "
                    f"got {val!r}"
                )
        elif not isinstance(val, bool):
            raise TypeError(
                f"dataclass option `{name}` must be of type `bool`. Got {val} "
                f"(type: {type(val)})"
            )
        merged[name] = val
    if (
        include_module
        and "module" not in merged
        and "module" in _STRICT_DATACLASS_OPTION_KEYS
    ):  # pragma: no cover
        # For Python 3.12+ we want the default module to
        # remain "types" rather than being inferred as some
        # internal hydra-zen module.
        merged["module"] = "types"
    return merged


def parse_strict_dataclass_options(
    options: Mapping[str, Any]
) -> TypeGuard[StrictDataclassOptions]:
    return (
        options.keys() <= _STRICT_DATACLASS_OPTION_KEYS
        and StrictDataclassOptions.__required_keys__ <= options.keys()
    )


_HYDRA_CONVERT_OPTIONS = (
    {"none", "partial", "all", "object"}
    if HYDRA_SUPPORTS_OBJECT_CONVERT
    else {"none", "partial", "all"}
)


def validate_hydra_options(
    hydra_recursive: Optional[bool] = None,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = None,
) -> None:
    if hydra_recursive is not None and not isinstance(hydra_recursive, bool):
        raise TypeError(
            f"`hydra_recursive` must be a boolean type, got {hydra_recursive}"
        )

    if hydra_convert is not None and hydra_convert not in _HYDRA_CONVERT_OPTIONS:
        raise ValueError(
            f"`hydra_convert` must be 'none', 'partial',"
            f"{' object' if HYDRA_SUPPORTS_OBJECT_CONVERT else ''} or 'all', got: "
            f"{hydra_convert}"
        )



# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from collections import Counter, defaultdict, deque
from datetime import timedelta
from enum import Enum
from functools import partial
from pathlib import Path, PosixPath, WindowsPath
from typing import Final, NamedTuple

import hydra
import omegaconf

NoneType = type(None)


class Version(NamedTuple):
    major: int
    minor: int
    patch: int


def _get_version(ver_str: str) -> Version:
    # Not for general use. Tested only for Hydra and OmegaConf
    # version string styles

    splits = ver_str.split(".")[:3]
    if not len(splits) == 3:  # pragma: no cover
        raise ValueError(f"Version string {ver_str} couldn't be parsed")

    major, minor = (int(v) for v in splits[:2])
    patch_str, *_ = splits[-1].split("rc")

    return Version(major=major, minor=minor, patch=int(patch_str))


OMEGACONF_VERSION: Final = _get_version(omegaconf.__version__)
HYDRA_VERSION: Final = _get_version(hydra.__version__)


# Indicates primitive types permitted in type-hints of structured configs
HYDRA_SUPPORTED_PRIMITIVE_TYPES: Final = frozenset(
    {int, float, bool, str, Enum, bytes, Path}
)
# Indicates types of primitive values permitted in configs
HYDRA_SUPPORTED_PRIMITIVES = frozenset(
    {
        int,
        float,
        bool,
        str,
        list,
        tuple,
        dict,
        NoneType,
        bytes,
        Path,
        PosixPath,
        WindowsPath,
    }
)
ZEN_SUPPORTED_PRIMITIVES: frozenset[type] = frozenset(
    {
        set,
        frozenset,
        complex,
        partial,
        bytearray,
        deque,
        Counter,
        range,
        timedelta,
        defaultdict,
    }
)


HYDRA_SUPPORTS_OBJECT_CONVERT = HYDRA_VERSION >= Version(1, 3, 0)



# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Simple helper functions used to implement `just` and `builds`. This module is designed specifically so
that these functions have a legible module-path when they appear in configuration files.
"""
import functools as _functools
import typing as _tp
from functools import partial
from typing import Any, Callable, TypeVar

from hydra._internal import utils as _hydra_internal_utils
from hydra.utils import log as _log
from typing_extensions import TypeAlias

from hydra_zen.structured_configs._utils import (
    is_interpolated_string as _is_interpolated_string,
)
from hydra_zen.typing import Partial as _Partial

__all__ = ["partial", "get_obj", "zen_processing"]

_T = _tp.TypeVar("_T")

_Wrapper = _tp.Callable[[_tp.Callable[..., _tp.Any]], _tp.Callable[..., _tp.Any]]
_WrapperConf = _tp.Union[str, _Wrapper]


T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])

Wrappers: TypeAlias = tuple[Callable[[Callable[..., Any]], Callable[..., Any]], ...]


class partial_with_wrapper(partial[T]):
    """`partial` where a wrapper's application is deferred until the call."""

    __slots__ = "wrappers"
    wrappers: Wrappers

    def __new__(
        cls,
        wrappers: Wrappers,
        func: Callable[..., T],
        *args: Any,
        **keywords: Any,
    ) -> "partial_with_wrapper[T]":
        if isinstance(func, partial_with_wrapper):
            # unwrap the partial_with_wrapper
            wrappers = func.wrappers + wrappers
            keywords = {**func.keywords, **keywords}
            args = func.args + args
            func = func.func

        self = super().__new__(cls, func, *args, **keywords)
        self.wrappers = wrappers
        return self

    def __call__(self, /, *args: Any, **keywords: Any) -> T:
        keywords = {**self.keywords, **keywords}
        func = self.func
        for wrapper in self.wrappers:
            func = wrapper(func)

        return func(*self.args, *args, **keywords)

    def __reduce__(self) -> _tp.Any:
        return (
            type(self),
            (
                self.wrappers,
                self.func,
            ),
            (
                self.wrappers,
                self.func,
                self.args,
                self.keywords or None,
                self.__dict__ or None,
            ),
        )

    def __setstate__(
        self,
        state: tuple[
            tuple[Callable[..., Any], ...],
            Callable[..., Any],
            tuple[Any, ...],
            dict[str, Any],
            dict[str, Any],
        ],
    ) -> None:
        if not isinstance(state, tuple):  # pragma: no cover
            raise TypeError("argument to __setstate__ must be a tuple")
        if len(state) != 5:  # pragma: no cover
            raise TypeError(f"expected 4 items in state, got {len(state)}")

        self.wrappers, *substate = state
        substate = tuple(substate)
        super().__setstate__(substate)  # type: ignore


def partial(
    *args: _tp.Any,
    _partial_target_: _tp.Callable[..., _T],
    **kwargs: _tp.Any,
) -> _Partial[_T]:
    """Equivalent to ``functools.partial`` but provides a named parameter for the callable."""
    return _tp.cast(_Partial[_T], _functools.partial(_partial_target_, *args, **kwargs))


def get_obj(*, path: str) -> _tp.Union[type, _tp.Callable[..., _tp.Any]]:
    """Imports an object given the specified path."""
    try:
        cl = _hydra_internal_utils._locate(path)
        return cl
    except Exception as e:  # pragma: no cover
        _log.error(f"Error getting callable at {path} : {e}")
        raise e


def zen_processing(
    *args: _tp.Any,
    _zen_target: str,
    _zen_partial: bool = False,
    _zen_exclude: _tp.Sequence[str] = tuple(),
    _zen_wrappers: _tp.Union[_WrapperConf, _tp.Sequence[_WrapperConf]] = tuple(),
    _zen_target_wrapper: _tp.Union[None, _Wrapper] = None,
    **kwargs: _tp.Any,
) -> _tp.Any:

    if isinstance(_zen_wrappers, str) or not isinstance(_zen_wrappers, _tp.Sequence):
        unresolved_wrappers: _tp.Sequence[_WrapperConf] = (_zen_wrappers,)
    else:
        unresolved_wrappers: _tp.Sequence[_WrapperConf] = _zen_wrappers
    del _zen_wrappers

    resolved_wrappers: _tp.List[_Wrapper] = []

    for _unresolved in unresolved_wrappers:
        if _unresolved is None:
            # We permit interpolated fields to resolve to `None`; this is
            # a nice pattern for enabling people to ergonomically toggle
            # wrappers off.
            continue
        if isinstance(_unresolved, str):
            # Hydra will have already raised on missing interpolation
            # keys by here
            assert not _is_interpolated_string(_unresolved)
            _unresolved = get_obj(path=_unresolved)

        if not callable(_unresolved):
            raise TypeError(
                f"Instantiating {_zen_target}: `zen_wrappers` was passed a non-callable object: {_unresolved}"
            )
        else:
            resolved = _unresolved
        del _unresolved
        resolved_wrappers.append(resolved)

    obj = get_obj(path=_zen_target)
    if _zen_target_wrapper is not None:
        resolved_wrappers = [_zen_target_wrapper] + resolved_wrappers

    # first wrapper listed should be called first
    # [f1, f2, f3, ...] ->
    #    target = f1(target)
    #    target = f2(target)
    #    target = f3(target)
    #    ...

    if _zen_exclude:
        excluded_set = set(_zen_exclude)
        kwargs = {k: v for k, v in kwargs.items() if k not in excluded_set}

    if _zen_partial is True:
        if not resolved_wrappers:
            return _functools.partial(obj, *args, **kwargs)
        else:
            # if we have wrappers, we need to use the partial_with_wrapper
            # class to ensure that the wrappers are called in the right order
            return partial_with_wrapper(tuple(resolved_wrappers), obj, *args, **kwargs)
    for wrapper in resolved_wrappers:
        obj = wrapper(obj)
    return obj(*args, **kwargs)


def as_default_dict(
    dict_: _tp.Dict[_tp.Any, _tp.Any], *, default_factory: _tp.Any
) -> _tp.DefaultDict[_tp.Any, _tp.Any]:
    from collections import defaultdict

    obj = defaultdict(default_factory)
    obj.update(dict_)
    return obj



# Copyright (c) 2025 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from typing import Optional, Union

from typing_extensions import Literal

from hydra_zen.typing import DataclassOptions, SupportedPrimitive
from hydra_zen.typing._implementations import (
    AllConvert,
    DataClass,
    DataClass_,
    DefaultsList,
    ZenConvert,
)

from ._implementations import DefaultBuilds, ZenField

__all__ = ["ZenField", "make_config"]


class NOTHING:
    def __init__(self) -> None:
        raise TypeError("`NOTHING` cannot be instantiated")


_MAKE_CONFIG_SETTINGS = AllConvert(dataclass=False, flat_target=False)


def make_config(
    *fields_as_args: Union[str, ZenField],
    hydra_recursive: Optional[bool] = None,
    hydra_convert: Optional[Literal["none", "partial", "all", "object"]] = None,
    hydra_defaults: Optional[DefaultsList] = None,
    zen_dataclass: Optional[DataclassOptions] = None,
    bases: tuple[type[DataClass_], ...] = (),
    zen_convert: Optional[ZenConvert] = None,
    **fields_as_kwargs: Union[SupportedPrimitive, ZenField],
) -> type[DataClass]:
    """
    Returns a config with user-defined field names and, optionally,
    associated default values and/or type annotations.

    Unlike `builds`, `make_config` is not used to configure a particular target
    object; rather, it can be used to create more general configs [1]_.

    Parameters
    ----------
    *fields_as_args : str | ZenField
        The names of the fields to be be included in the config. Or, `ZenField`
        instances, each of which details the name and their default value and/or the
        type annotation of a given field.

    **fields_as_kwargs : SupportedPrimitive | ZenField
        Like ``fields_as_args``, but field-name/default-value pairs are
        specified as keyword arguments. `ZenField` can also be used here
        to express a field's type-annotation and/or its default value.

        Named parameters of the forms ``hydra_xx``, ``zen_xx``, and ``_zen_xx`` are reserved to ensure future-compatibility, and cannot be specified by the user.

    zen_convert : Optional[ZenConvert]
        A dictionary that modifies hydra-zen's value and type conversion behavior.
        Consists of the following optional key-value pairs (:ref:`zen-convert`):

        - `dataclass` : `bool` (default=False):
            If `True` any dataclass type/instance without a
            `_target_` field is automatically converted to a targeted config
            that will instantiate to that type/instance. Otherwise the dataclass
            type/instance will be passed through as-is.

    bases : Tuple[Type[DataClass], ...], optional (default=())
        Base classes that the resulting config class will inherit from.

    hydra_recursive : Optional[bool], optional (default=True)
        If ``True``, then Hydra will recursively instantiate all other
        hydra-config objects nested within this dataclass [2]_.

        If ``None``, the ``_recursive_`` attribute is not set on the resulting config.

    hydra_convert : Optional[Literal["none", "partial", "all", "object"]], optional (default="none")
        Determines how Hydra handles the non-primitive objects passed to configuration [3]_.

        - ``"none"``: Passed objects are DictConfig and ListConfig, default
        - ``"partial"``: Passed objects are converted to dict and list, with the exception of Structured Configs (and their fields).
        - ``"all"``: Passed objects are dicts, lists and primitives without a trace of OmegaConf containers
        - ``"object"``: Passed objects are converted to dict and list. Structured Configs are converted to instances of the backing dataclass / attr class.

        If ``None``, the ``_convert_`` attribute is not set on the resulting config.

    hydra_defaults : None | list[str | dict[str, str | list[str] | None ]], optional (default = None)
        A list in an input config that instructs Hydra how to build the output config
        [7]_ [8]_. Each input config can have a Defaults List as a top level element. The
        Defaults List itself is not a part of output config.

    zen_dataclass : Optional[DataclassOptions]
        A dictionary can specify any option that is supported by
        :py:func:`dataclasses.make_dataclass` other than `fields`.

        Additionally, a ``'module': <str>`` entry can be specified to enable pickle
        compatibility. See `hydra_zen.typing.DataclassOptions` for details.

    frozen : bool, optional (default=False)
        .. deprecated:: 0.9.0
            `frozen` will be removed in hydra-zen 0.10.0. It is replaced by
            ``zen_dataclass={'frozen': <bool>}``.

        If ``True``, the resulting config class will produce 'frozen' (i.e. immutable)
        instances. I.e. setting/deleting an attribute of an instance of the config will
        raise :py:class:`dataclasses.FrozenInstanceError` at runtime.

    config_name : str, optional (default="Config")
        .. deprecated:: 0.9.0
            `config_name` will be removed in hydra-zen 0.10.0. It is replaced by
            ``zen_dataclass={'cls_name': <str>}``.

        The class name of the resulting config class.

    Returns
    -------
    Config : Type[DataClass]
        The resulting config class; a dataclass that possess the user-specified fields.

    Raises
    ------
    hydra_zen.errors.HydraZenUnsupportedPrimitiveError
        The provided configured value cannot be serialized by Hydra, nor does hydra-zen
        provide specialized support for it. See :ref:`valid-types` for more details.

    Notes
    -----
    The resulting "config" is a dataclass-object [4]_ with Hydra-specific attributes
    attached to it, along with the attributes specified via ``fields_as_args`` and
    ``fields_as_kwargs``. **Unlike std-lib dataclasses, the default value for
    unsafe_hash is True.**

    Any field specified without a type-annotation is automatically annotated with
    :py:class:`typing.Any`. Hydra only supports a narrow subset of types [5]_;
    `make_config` will automatically 'broaden' any user-specified annotations so that
    they are compatible with Hydra.

    `make_config` will automatically manipulate certain types of default values to
    ensure that they can be utilized in the resulting config and by Hydra:

    - Mutable default values will automatically be packaged in a default factory function [6]_
    - A default value that is a class-object or function-object will automatically be wrapped by `just`, to ensure that the resulting config is serializable by Hydra.

    For finer-grain control over how type annotations and default values are managed,
    consider using :func:`dataclasses.make_dataclass`.

    For details of the annotation `SupportedPrimitive`, see :ref:`valid-types`.

    See Also
    --------
    builds : Create a targeted structured config designed to "build" a particular object.
    just : Create a config that "just" returns a class-object or function, without instantiating/calling it.

    References
    ----------
    .. [1] https://hydra.cc/docs/tutorials/structured_config/intro/
    .. [2] https://hydra.cc/docs/advanced/instantiate_objects/overview/#recursive-instantiation
    .. [3] https://hydra.cc/docs/advanced/instantiate_objects/overview/#parameter-conversion-strategies
    .. [4] https://docs.python.org/3/library/dataclasses.html
    .. [5] https://hydra.cc/docs/tutorials/structured_config/intro/#structured-configs-supports
    .. [6] https://docs.python.org/3/library/dataclasses.html#default-factory-functions
    .. [7] https://hydra.cc/docs/tutorials/structured_config/defaults/
    .. [8] https://hydra.cc/docs/advanced/defaults_list/

    Examples
    --------
    >>> from hydra_zen import make_config, to_yaml
    >>> def pp(x):
    ...     return print(to_yaml(x))  # pretty-print config as yaml

    **Basic Usage**

    Let's create a bare-bones config with two fields, named 'a' and 'b'.

    >>> Conf1 = make_config("a", "b")  # sig: `Conf(a: Any, b: Any)`
    >>> pp(Conf1)
    a: ???
    b: ???

    Now we'll configure these fields with particular values:

    >>> conf1 = Conf1(1, "hi")
    >>> pp(conf1)
    a: 1
    b: hi
    >>> conf1.a
    1
    >>> conf1.b
    'hi'

    We can also specify fields via keyword args; this is especially convenient
    for providing associated default values.

    >>> Conf2 = make_config("unit", data=[-10, -20])
    >>> pp(Conf2)
    unit: ???
    data:
    - -10
    - -20

    Configurations can be nested

    >>> Conf3 = make_config(c1=Conf1(a=1, b=2), c2=Conf2)
    >>> pp(Conf3)
    c1:
      a: 1
      b: 2
    c2:
      unit: ???
      data:
      - -10
      - -20
    >>> Conf3().c1.a
    1

    Configurations can be composed via inheritance

    >>> ConfInherit = make_config(c=2, bases=(Conf2, Conf1))
    >>> pp(ConfInherit)
    a: ???
    b: ???
    unit: ???
    data:
    - -10
    - -20
    c: 2

    >>> issubclass(ConfInherit, Conf1) and issubclass(ConfInherit, Conf2)  # type: ignore
    True

    **Support for Additional Types**

    Types like :py:class:`complex` and :py:class:`pathlib.Path` are automatically
    supported by hydra-zen.

    >>> ConfWithComplex = make_config(a=1+2j)
    >>> pp(ConfWithComplex)
    a:
      real: 1.0
      imag: 2.0
      _target_: builtins.complex

    See :ref:`additional-types` for a complete list of supported types.

    **Using ZenField to Provide Type Information**

    The `ZenField` class can be used to include a type-annotation in association
    with a field.

    >>> from hydra_zen import ZenField as zf
    >>> ProfileConf = make_config(username=zf(str), age=zf(int))
    >>> # signature: ProfileConf(username: str, age: int)

    Providing type annotations is optional, but doing so enables Hydra to perform
    checks at runtime to ensure that a configured value matches its associated
    type [4]_.

    >>> pp(ProfileConf(username="piro", age=False))  # age should be an integer
    <ValidationError: Value 'False' could not be converted to Integer>

    These default values can be provided alongside type annotations

    >>> C = make_config(age=zf(int, 0))  # signature: C(age: int = 0)

    `ZenField` can also be used to specify ``fields_as_args``; here, field names
    must be specified as well.

    >>> C2 = make_config(zf(name="username", hint=str), age=zf(int, 0))
    >>> # signature: C2(username: str, age: int = 0)

    See :ref:`data-val` for more general data validation capabilities via hydra-zen.
    """
    _locals = locals().copy()
    fields_as_args = _locals.pop("fields_as_args")
    fields_as_kwargs = _locals.pop("fields_as_kwargs")
    return DefaultBuilds.make_config(*fields_as_args, **_locals, **fields_as_kwargs)  # type: ignore
